from typing import List
from pydantic import BaseModel, Field, ConfigDict


class Message(BaseModel):
    """This class is a wrapper around a message. It is a building block of a history.

    Attributes:
        role (str): The role of the message. Something like "user" or "system".
        content (str): The content of the message. A string.
    """

    model_config = ConfigDict(extra="allow")

    role: str = Field(
        ...,
        description="The role of the message. Something like 'user' or 'system'.",
    )
    content: str = Field(
        ..., description="The content of the message. A string."
    )


class History:
    """
    This class is a wrapper around a list of messages. It is used to keep track of the conversation history with an LLM.

    A history is a list of messages. Each message has a role and content. The role is something like "user" or "system". The content is a string.
    """

    def __init__(self, system_message: str = None):
        """Constructor for the History class.

        Args:
            system_message (str, optional): System message for the LLM to use, if any. E.g., "You are a helpful assistant." Defaults to None.
        """
        self._value: List[Message] = []
        self._cost = 0.0
        if system_message:
            self.add(Message(role="system", content=system_message))

    @property
    def cost(self):
        return self._cost

    def increment_cost(self, cost: float):
        """Increments the cost of the history.

        Args:
            cost (float): Float representing the cost to increment by (in dollars)
        """
        self._cost += cost

    def __getitem__(self, index: int):
        """Returns the message at the given index."""
        return self._value[index]

    def __len__(self):
        """Returns the number of messages in the history."""
        return len(self._value)

    def add(self, message: Message):
        """Adds a message to the history. Used in LLM run methods."""

        self._value.append(message)

    def __iter__(self):
        """An iterator over the messages in the history.

        Usage:
        ```python
        h = History()

        for message in h:
            print(message)
        ```

        Returns:
            _type_: _description_
        """
        return iter(self._value)

    def __str__(self):
        """String representation of the history."""
        messages_str = "\n".join(
            [
                f"[{message.role.capitalize()}]: {message.content}"
                for message in self._value
            ]
        )
        return f"History:\n{messages_str}"



from abc import ABC, abstractmethod

from gears.history import History


class BaseLLM(ABC):
    @abstractmethod
    async def run(
        prompt: str,
        history: History,
        message_kwargs: dict = {},
    ):
        pass



"""
This file contains the OpenAI chat API wrappers.
"""
from gears.history import History, Message
from gears.llms.base import BaseLLM
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
)  # for exponential backoff
import openai
from typing import Any

import logging

logger = logging.getLogger(__name__)


def retry_after_attempts(max_retries: int):
    return retry(
        wait=wait_random_exponential(min=1, max=60),
        stop=stop_after_attempt(max_retries),
    )


OPENAI_PRICING_MAP = {
    "gpt-3.5-turbo": {
        "prompt_tokens": float(0.0015 / 1000),
        "completion_tokens": float(0.002 / 1000),
    },
    "gpt-3.5-turbo-16k": {
        "prompt_tokens": float(0.003 / 1000),
        "completion_tokens": float(0.004 / 1000),
    },
    "gpt-4": {
        "prompt_tokens": float(0.03 / 1000),
        "completion_tokens": float(0.06 / 1000),
    },
    "gpt-4-32k": {
        "prompt_tokens": float(0.06 / 1000),
        "completion_tokens": float(0.12 / 1000),
    },
}


class OpenAIChat(BaseLLM):
    def __init__(
        self, model: str = "gpt-3.5-turbo", max_retries: int = 3, **kwargs
    ):
        self.model = model
        self.api_kwargs = kwargs
        self.max_retries = max_retries

        # strip date from model
        split_model = self.model.split("-")
        if len(split_model) >= 4 and "k" not in split_model[-1]:
            # Truncade the last part of the model name
            self.model_base = "-".join(split_model[:-1])
        else:
            self.model_base = self.model

    def retry_policy(self):
        return retry(
            wait=wait_random_exponential(min=1, max=60),
            stop=stop_after_attempt(self.max_retries),
        )

    async def chat_api_call(self, request: dict):
        retry_decorator = self.retry_policy()
        decorated_func = retry_decorator(self._chat_api_call_impl)
        return await decorated_func(request)

    async def _chat_api_call_impl(self, request: dict):
        try:
            response = await openai.ChatCompletion.acreate(**request)
            return response
        except Exception as e:
            logger.warning(
                f"Exception when calling OpenAI occurred: {e}. Retrying..."
            )
            raise

    async def run(
        self,
        prompt: str,
        history: History,
        message_kwargs: dict = {},
    ) -> Any:
        # Construct chat history
        curr_message = Message(role="user", content=prompt, **message_kwargs)
        history.add(curr_message)
        messages = [m.model_dump() for m in history]
        request = {
            "model": self.model,
            "messages": messages,
            **self.api_kwargs,
        }
        response = await self.chat_api_call(request)
        returned_message = response["choices"][0]["message"]["content"]
        history.add(
            Message(
                role=response["choices"][0]["message"]["role"],
                content=returned_message,
            )
        )

        # Increment cost
        try:
            prompt_tokens = response["usage"]["prompt_tokens"]
            completion_tokens = response["usage"]["completion_tokens"]

            cost = (
                OPENAI_PRICING_MAP[self.model_base]["prompt_tokens"]
                * prompt_tokens
                + OPENAI_PRICING_MAP[self.model_base]["completion_tokens"]
                * completion_tokens
            )
            history.increment_cost(cost)
        except KeyError:
            logger.error(
                f"Could not find pricing for model {self.model}. Not incrementing cost."
            )

        return response



import json

from typing import Generator, List, Dict, Union


def extract_json(
    text, decoder=json.JSONDecoder()
) -> Generator[Union[Dict, List], None, None]:
    """
    Generates all JSON objects or arrays in a string representing the output of an LLM. Copied from ChatGPT.

    Args:
        text (str): The string to parse.

    Returns:
        Generator[Union[Dict, List], None, None]: A generator of JSON objects or arrays.
    """
    pos = 0
    while True:
        match = text.find("{", pos)
        match2 = text.find("[", pos)
        # Find the earliest opening bracket
        if (match != -1 and match < match2) or match2 == -1:
            start_pos = match
        else:
            start_pos = match2

        if start_pos == -1:
            break
        try:
            result, index = decoder.raw_decode(text[start_pos:])
            yield result
            pos = start_pos + index
        except json.JSONDecodeError:
            pos = start_pos + 1


def extract_first_json(text: str) -> Union[Dict, List]:
    """
    Extracts the first JSON object or array in a string representing the output of an LLM.

    Args:
        text (str): The string to parse.

    Returns:
        Union[Dict, List]: The first JSON object or array.
    """
    return next(extract_json(text))
